from __future__ import annotations

import logging

import js

from wwwpy.common.designer.html_edit import Position
from wwwpy.common.designer.ui._drop_indicator_svg import svg_indicator_for
from wwwpy.common.designer.ui.rect_readonly import RectReadOnly
from wwwpy.remote import dict_to_js
from wwwpy.remote.designer.ui.floater import Floater

logger = logging.getLogger(__name__)


class DropIndicatorFloater(Floater, tag_name='wwwpy-drop-indicator-floater'):

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
        <style>
            :host {
              position: fixed;
              pointer-events: none;
              //border: 2px solid #4a90e2;
              z-index: 200000;
              display: none;
            } 
        </style>      
        """

    @property
    def transition(self) -> bool:
        return self.element.style.transition != 'none'

    @transition.setter
    def transition(self, value: bool):
        if value:
            self.element.style.transition = 'all 0.2s ease'
        else:
            self.element.style.transition = 'none'

    @property
    def visible(self) -> bool:
        return self.element.style.display == 'block'

    def hide(self):
        self.element.style.display = 'none'

    def show(self):
        self.element.style.display = 'block'

    def set_reference_geometry(self, rect: RectReadOnly):
        raise NotImplementedError()

    def set_reference_geometry2(self, rect: RectReadOnly, position: Position):

        bs = 2  # Adjust this value to match the border size in CSS

        r = js.DOMRect.new(rect.x - bs, rect.y - bs, rect.width, rect.height, )

        self.element.style.display = 'block'
        self.element.style.top = f"{r.top}px"
        self.element.style.left = f"{r.left}px"
        self.element.style.width = f"{r.width}px"
        self.element.style.height = f"{r.height}px"

        sr = self.element.shadowRoot
        while sr.children.length > 1:  # leave style element
            sr.children[1].remove()

        svg = svg_indicator_for(r.width + bs * 2, r.height + bs * 2, position)
        fragment = js.document.createRange().createContextualFragment(svg)
        sr.appendChild(fragment)


def compute_xy(W, H):
    import math
    k = (1 - 1 / math.sqrt(3)) / 2
    return round(W * k, 9), round(H * k)


def create_svg(w, h, x, y, pulse_option) -> str:
    """Create SVG string based on parameters."""
    # Validate to ensure inner rectangle has positive dimensions
    if x * 2 >= w or y * 2 >= h:
        return f'<text x="10" y="30" fill="red">Error: Inset values too large for the given width/height</text>'

    # Calculate inner rectangle dimensions
    inner_width = w - 2 * x
    inner_height = h - 2 * y

    # Determine which parts should pulse
    top_class, left_class, bottom_class, right_class, inner_class = [''] * 5

    pulse = 'class="pulse" stroke="cyan"'
    if pulse_option == "top_left":
        top_class, left_class = pulse, pulse
    elif pulse_option == "bottom_right":
        bottom_class, right_class = pulse, pulse
    elif pulse_option == "inner":
        inner_class = pulse

    # Create SVG string with the CSS classes for animation
    # return _svg_at_begin
    return _svg_at_end


# language=html
_svg_at_begin = f"""
<svg style='width: 200px ; height: 200px;' viewBox="0 0 100 100" xmlns="http://www.w3.org/2000/svg">
    <g>
        <!-- Main diagonal line (arrow body) - shorter -->
        <line x1="70" y1="70" x2="0" y2="0" stroke-width="6" stroke="black">
            <animate
                    attributeName="stroke"
                    values="black;white;black"
                    dur="1.5s"
                    repeatCount="indefinite"/>
        </line>

        <!-- First arrowhead segment -->
        <line x1="-3" y1="0" x2="47" y2="0" stroke-width="6" stroke="black">
            <animate
                    attributeName="stroke"
                    values="black;white;black"
                    dur="1.5s"
                    repeatCount="indefinite"/>
        </line>

        <!-- Second arrowhead segment -->
        <line x1="0" y1="-3" x2="0" y2="47" stroke-width="6" stroke="black">
            <animate
                    attributeName="stroke"
                    values="black;white;black"
                    dur="1.5s"
                    repeatCount="indefinite"/>
        </line>
        <animateTransform
                attributeName="transform"
                type="translate"
                values="3,3; 10,10; 3,3"
                dur="1s"
                repeatCount="indefinite"
                additive="sum"/>
    </g>
</svg>
"""
# language=html
_svg_at_end = f"""

<svg  style='width: 200px ; height: 200px;' viewBox="0 0 100 100" xmlns="http://www.w3.org/2000/svg">
    <g>
        <!-- Main diagonal line (arrow body) - shorter -->
        <line x1="30" y1="30" x2="100" y2="100" stroke-width="6" stroke="black">
            <animate
                    attributeName="stroke"
                    values="black;white;black"
                    dur="1.5s"
                    repeatCount="indefinite"/>
        </line>

        <!-- First arrowhead segment (horizontal) -->
        <line x1="53" y1="100" x2="103" y2="100" stroke-width="6" stroke="black">
            <animate
                    attributeName="stroke"
                    values="black;white;black"
                    dur="1.5s"
                    repeatCount="indefinite"/>
        </line>

        <!-- Second arrowhead segment (vertical) -->
        <line x1="100" y1="53" x2="100" y2="103" stroke-width="6" stroke="black">
            <animate
                    attributeName="stroke"
                    values="black;white;black"
                    dur="1.5s"
                    repeatCount="indefinite"/>
        </line>
        <animateTransform
                attributeName="transform"
                type="translate"
                values="-3,-3; -10,-10; -3,-3"
                dur="1s"
                repeatCount="indefinite"
                additive="sum"/>
    </g>
</svg>
"""



from __future__ import annotations

import js

from wwwpy.common.designer.ui.rect_readonly import RectReadOnly
from wwwpy.remote import dict_to_js
from wwwpy.remote.designer.ui.floater import Floater


class SelectionIndicatorFloater(Floater, tag_name='selection-indicator-floater'):

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
        <style>
            :host {
              position: fixed;
              pointer-events: none;
              border: 2px solid #4a90e2;
              background-color: rgba(74, 144, 226, 0.1);
              z-index: 200000;
              display: none;
            } 
        </style>      
        """

    @property
    def transition(self) -> bool:
        return self.element.style.transition != 'none'

    @transition.setter
    def transition(self, value: bool):
        if value:
            self.element.style.transition = 'all 0.2s ease'
        else:
            self.element.style.transition = 'none'

    @property
    def visible(self) -> bool:
        return self.element.style.display == 'block'

    def hide(self):
        self.element.style.display = 'none'

    def set_reference_geometry(self, rect: RectReadOnly):
        bs = 2  # Adjust this value to match the border size in CSS

        rect = js.DOMRect.new(rect.x - bs, rect.y - bs, rect.width, rect.height, )

        self.element.style.display = 'block'
        self.element.style.top = f"{rect.top}px"
        self.element.style.left = f"{rect.left}px"
        self.element.style.width = f"{rect.width}px"
        self.element.style.height = f"{rect.height}px"



from __future__ import annotations

import logging

import js
from pyodide.ffi import create_proxy

from wwwpy.common.designer.ui.rect_readonly import RectReadOnly
from wwwpy.common.injectorlib import injector
from wwwpy.remote import dict_to_js
from wwwpy.remote.designer.ui.design_aware import DesignAware
from wwwpy.remote.designer.ui.floater import Floater
from wwwpy.remote.designer.ui.locator_event import LocatorEvent
from wwwpy.remote.jslib import is_contained

logger = logging.getLogger(__name__)


class _DesignAware(DesignAware):

    def is_selectable_le(self, locator_event: LocatorEvent) -> bool | None:
        def is_container(element: js.Element) -> bool:
            return element.tagName.lower() == ActionBandFloater.component_metadata.tag_name.lower()

        is_cont = is_contained(locator_event.main_element, is_container)
        logger.debug(f'is_selectable_le is_contained inside ActionBandFloater:{is_cont} ')
        if is_cont:
            return False
        return None


_design_aware = _DesignAware()


def extension_point_register():
    DesignAware.EP_REGISTRY.register(_design_aware)


def extension_point_unregister():
    DesignAware.EP_REGISTRY.unregister(_design_aware)


class ActionBandFloater(Floater, tag_name='action-band-floater'):
    """A component for creating a toolbar button with an icon and label.
    Converted from the JavaScript implementation in selection-scroll-1.html.
    """

    def init_component(self):
        """Initialize the component"""
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
        <style>
             :host {
              position: fixed;
              display: none;
              background-color: #333;
              border-radius: 4px;
              padding: 4px;
              z-index: 200001;
              box-shadow: 0 2px 5px rgba(0,0,0,0.2);
              white-space: nowrap;
              min-width: max-content;
              pointer-events: auto;
            }

            button {
              background-color: transparent;
              color: white;
              border: none;
              width: 30px;
              height: 30px;
              border-radius: 3px;
              cursor: pointer;
              margin: 0 2px;
            }

            button:hover {
              background-color: rgba(255,255,255,0.2);
            }
        </style>
        """
        self._toolbar_dimensions = None
        self.toolbar_element = self.element

        button_data = [
            {'label': 'Parent', 'icon': '←'},
            {'label': 'Move up', 'icon': '↑'},
            {'label': 'Move down', 'icon': '↓'},
            {'label': 'Edit', 'icon': '✏️'},
            {'label': 'Delete', 'icon': '🗑️', 'function': _delete_element},
        ]

        for data in button_data:
            button = js.document.createElement('button')
            button.setAttribute('aria-label', data['label'])
            button.setAttribute('title', data['label'])
            button.textContent = data['icon']

            # Create a proxy function to handle the click event
            def create_button_click_handler(action_label, data=data):
                def button_click_handler(e):
                    e.stopPropagation()
                    # Dispatch a custom event when a toolbar button is clicked
                    self.element.dispatchEvent(
                        js.CustomEvent.new('toolbar-action', dict_to_js({
                            'bubbles': True,
                            'composed': True,
                            'detail': {
                                'action': action_label,
                            }
                        }))
                    )
                    fun = data.get('function', None)
                    if fun:
                        fun()
                    else:
                        js.alert(f"Button clicked: {data['label']}")

                return button_click_handler

            # Add event listener with the proxy function
            button.addEventListener('click', create_proxy(create_button_click_handler(data['label'])))
            self.element.shadowRoot.appendChild(button)

    def _handle_click(self, event):
        event.stopPropagation()

    def hide(self):
        self.element.style.display = 'none'

    def set_reference_geometry(self, rect: RectReadOnly):
        # self._toolbar_dimensions = None
        # Only measure the toolbar once initially to avoid layout thrashing
        if not self._toolbar_dimensions:
            self.toolbar_element.style.display = 'block'
            self.toolbar_element.style.visibility = 'hidden'
            self.toolbar_element.style.top = '-9999px'  # Position off-screen for measurement

            # Cache toolbar dimensions
            self._toolbar_dimensions = {
                'width': self.toolbar_element.offsetWidth,
                'height': self.toolbar_element.offsetHeight
            }

        # Use cached dimensions
        toolbar_width = self._toolbar_dimensions['width']
        toolbar_height = self._toolbar_dimensions['height']

        # Position the toolbar at bottom-right
        toolbar_x = rect.right - toolbar_width  # Align right edges
        toolbar_y = rect.bottom + 5  # Add 5px offset

        # Check if the toolbar would go off-screen to the left
        if toolbar_x < 0:
            toolbar_x = rect.left

        # Check if toolbar would go off-screen at the bottom
        if rect.bottom + toolbar_height + 5 > js.window.innerHeight:
            toolbar_y = rect.top - toolbar_height - 5

        self.toolbar_element.style.display = 'block'
        self.toolbar_element.style.visibility = 'visible'
        self.toolbar_element.style.left = f"{toolbar_x}px"
        self.toolbar_element.style.top = f"{toolbar_y}px"


def _delete_element():
    from wwwpy.common.asynclib import create_task_safe
    from wwwpy.common.designer.canvas_selection import CanvasSelection
    from wwwpy.common.designer.code_edit import remove_element
    from wwwpy.common.designer.html_locator import path_to_index
    canvas_selection = injector.get(CanvasSelection)
    el_path = canvas_selection.current_selection
    if not el_path:
        return
    from wwwpy.common import modlib
    file = modlib._find_module_path(el_path.class_module)
    old_source = file.read_text()

    path_index = path_to_index(el_path.path)
    remove_result = remove_element(old_source, el_path.class_name, path_index)
    if not remove_result:
        js.alert('Something went wrong')
        return

    logger.debug(f'write_module_file len={len(remove_result)} el_path={el_path}')

    async def write_source_file():
        from wwwpy.server.designer import rpc
        write_res = await rpc.write_module_file(el_path.class_module, remove_result)
        canvas_selection.current_selection = None
        logger.debug(f'write_module_file res={write_res}')

    create_task_safe((write_source_file()))


def _pretty(node: js.Element):
    if hasattr(node, 'tagName'):
        identifier = node.dataset.name if node.hasAttribute('data-name') else node.id
        return f'{node.tagName.lower()}#{identifier}.{node.className}[{node.innerHTML.strip()[:20]}…]'
    return str(node)



from __future__ import annotations

import asyncio
import logging
import random
from typing import TypeVar, Any, Type, TypeGuard, Callable, overload

import js
from js import document
from pyodide.ffi import create_once_callable
from pyodide.ffi import create_proxy

logger = logging.getLogger(__name__)


# logger.setLevel('DEBUG')


async def script_load_once(src: str, script_type='', **kwargs) -> bool:
    """This function loads a (javascript) script once. It waits for the script to load and returns True if the script was loaded,
    and False if the script was already loaded."""

    marker = f'm-{random.randint(1000, 9999)}'
    logger.debug(f'{marker} script_load_once: {src} {script_type} {kwargs}')

    scripts = list(filter(lambda s: s.src == src, document.head.querySelectorAll('script')))  # noqa
    scripts: list[js.HTMLScriptElement]
    need_load = not scripts
    script = None if need_load else scripts[0]

    if script:
        logger.debug(f'{marker}   found script with same src: {src} type={script.type}')
        assert script.type == script_type, f'script.type=`{script.type}`, script_type=`{script_type}`'
    else:
        logger.debug(f'{marker}  load is needed, no script with same src found: {src}')
        script = document.createElement('script')
        # copy all kwargs to the script
        for key, value in kwargs.items():
            setattr(script, key, value)
        script.src = src
        if script_type:
            script.type = script_type
        script.__async_event = asyncio.Event()
        script.onload = create_proxy(lambda e: script.__async_event.set())
        document.head.append(script)

    logger.debug(f'{marker}  waiting for script to load: {src}')
    await script.__async_event.wait()
    logger.debug(f'{marker}  event set: need_load={need_load} {src}')
    return need_load


async def waitAnimationFrame():
    event = asyncio.Event()
    js.window.requestAnimationFrame(create_once_callable(lambda *_: event.set()))
    await event.wait()


class AnimationFrameTracker:
    """Tracks animation frames and calls a callback."""

    def __init__(self, callback):
        self._callback = callback
        self._raf_id = None
        self._on_animation_frame = create_proxy(self._on_animation_frame)

    def start(self):
        if self._raf_id is None:
            self._raf_id = js.window.requestAnimationFrame(self._on_animation_frame)

    def stop(self):
        if self._raf_id is not None:
            js.window.cancelAnimationFrame(self._raf_id)
            self._raf_id = None

    def _on_animation_frame(self, timestamp):
        if self._raf_id is None:
            return

        self._raf_id = js.window.requestAnimationFrame(self._on_animation_frame)
        self._callback(timestamp)

    @property
    def is_tracking(self):
        return self._raf_id is not None


_instanceof: Callable[[..., ...], bool] = js.eval('(i,t) => i instanceof t')

T = TypeVar('T')


def is_instance_of(instance: Any, js_type: Type[T]) -> TypeGuard[T]:
    """Check if the instance is of the given JavaScript type.
    Example: is_instance_of(js.document, js.HTMLDocument) will return True.
    """
    return _instanceof(instance, js_type)


@overload
def is_contained(target: js.Element, container: js.Element):
    """Check if target is a descendant of container, including shadow DOM and slots."""


@overload
def is_contained(target: js.Element, is_container: Callable[[js.Element], bool]):
    """the second argument will be called with all the pertinent nodes in the DOM tree;
    if it returns True, the target is considered contained."""


def is_contained(target, container):
    """Determines if target is a descendant of container, accounting for shadow DOM and slots."""
    if target is None:
        raise ValueError(f'target is None')
    if container is None:
        raise ValueError(f'container is None')
    if callable(container):
        is_container = container
        logger.debug(f'target: `{_pretty(target)}` container: `lambda...`')
    else:
        is_container = lambda x: x == container
        logger.debug(f'target: `{_pretty(target)}` container: `{_pretty(container)}`')
    node = target
    while node:
        if hasattr(node, 'tagName') and is_container(node):
            return True

        # Check if node is assigned to a slot
        if hasattr(node, "assignedSlot") and node.assignedSlot:
            node = node.assignedSlot
            logger.debug(f'assignedSlot: {_pretty(node)}')
            continue

        # If in a shadow tree (i.e., root has a host), move to the host element.
        if is_instance_of(node, js.ShadowRoot):
            node = node.host
            logger.debug(f'host: {_pretty(node)}')
        else:
            # Otherwise, move up in the light DOM.
            if not hasattr(node, 'parentNode'):
                # logger.debug(f'node has no parentNode: {_pretty(node)}')
                # break
                raise ValueError(f'node has no parentNode: {_pretty(node)}')
            node = node.parentNode
            logger.debug(f'parentNode: {_pretty(node)}')

    return False


def get_deepest_element(x, y) -> js.Element | None:
    """
    Get the deepest ele at the event coordinates by recursively traversing shadow DOMs.
    """

    def _rec(root):
        ele = root.elementFromPoint(x, y)
        ele_shadow = _shadow_root_of(ele)
        if root == ele_shadow:
            return root.host

        if ele and ele_shadow:
            rec = _rec(ele_shadow)
            if rec:
                return rec

        return ele

    return _rec(js.document)


def _shadow_root_of(element) -> js.ShadowRoot | None:
    if element is None:
        return None

    if hasattr(element, 'shadowRoot') and element.shadowRoot and is_instance_of(element.shadowRoot, js.ShadowRoot):
        return element.shadowRoot
    return None


def _pretty(node: js.Element):
    if hasattr(node, 'tagName'):
        identifier = node.dataset.name if node.hasAttribute('data-name') else node.id
        return f'{node.tagName.lower()}#{identifier}.{node.className}[{node.innerHTML.strip()[:20]}…]'
    return str(node)


def closest_across_shadow(el, selector):
    current = el
    while current:
        match = current.closest(selector)
        if match:
            return match
        root = current.getRootNode()
        if not is_instance_of(root, js.ShadowRoot):
            return None
        current = root.host



import logging

import js

import wwwpy.remote.component as wpc

logger = logging.getLogger(__name__)


class SystemToolsComponent(wpc.Component, tag_name='wwwpy-system-tools'):
    _explore_fs: js.HTMLElement = wpc.element()
    _logger_levels: js.HTMLElement = wpc.element()
    _python_console: js.HTMLButtonElement = wpc.element()
    _system_versions: js.HTMLButtonElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<div style='display: inline-flex; flex-direction: column; gap: 5px' >        
<button data-name="_explore_fs">Explore Filesystem</button>
<button data-name="_logger_levels">Logger Levels</button>
<button data-name="_python_console">Python Console</button>
<button data-name="_system_versions">System Versions</button>
</div>
"""

    async def _explore_fs__click(self, event):
        from wwwpy.remote.designer.ui.dev_mode_component import DevModeComponent
        from wwwpy.remote.designer.ui import filesystem_tree
        filesystem_tree.show_explorer(DevModeComponent.instance.root_element())

    async def _logger_levels__click(self, event):
        from wwwpy.remote.designer.ui.dev_mode_component import DevModeComponent
        from wwwpy.remote.designer.ui.window_component import new_window
        w1 = new_window('Logger levels')

        from wwwpy.remote.designer.ui.system_tools.logger_levels import LoggerLevelsComponent
        w1.element.append(LoggerLevelsComponent().element)
        DevModeComponent.instance.root_element().append(w1.element)

    async def _python_console__click(self, event):
        from wwwpy.remote.designer.ui.dev_mode_component import DevModeComponent
        from wwwpy.remote.designer.ui.window_component import new_window
        w1 = new_window('Python console')
        from wwwpy.remote.designer.ui.python_console import PythonConsoleComponent
        w1.element.append(PythonConsoleComponent().element)
        DevModeComponent.instance.root_element().append(w1.element)

    async def _system_versions__click(self, event):
        from wwwpy.remote.designer.ui.dev_mode_component import DevModeComponent
        from wwwpy.remote.designer.ui.system_tools.system_versions import SystemVersions
        DevModeComponent.instance.show_window('Show System Versions', SystemVersions())
